﻿using UnityEngine;
using System.Collections;

public class playerController : MonoBehaviour 
{	
	// these code objects are paired with GUI objects in the Player's Inspector
	// click Player in the Hierarchy, open the Inspector, and voila!
	public GUIText countText;
	public GUIText winText;
	public GUIText velText;
	public GUIText velRollingText;
	public GUIText mhText;
	public GUIText mvText;

	// public variable will appear in Unity inspector for all objects with this attached script
	public float speed;

	// velocity of the ball
	private float pVel;

	// Unity event for wall impacts
	private FMOD.Studio.EventInstance wallHit;
	// parameter in FMOD "wall-hit-vel" event 
	private FMOD.Studio.ParameterInstance sndVel;
	// Unity event for ball movement
	private FMOD.Studio.EventInstance rollBall;
	// parameter in FMOD for "rolling" event 
	private FMOD.Studio.ParameterInstance rollVel;

	// variable to hold count value; incremented on each sucessful collection of a "Pickup" or collision with a cube
	private int count;

	// function executes when game begins
	void Start ()
	{
		// initialize count at 0
		count = 0;
		// call function to update GUIText field with count details
		SetCountText ();
		// set winning text field to be blank
		winText.text = "";

		// link the rolling sound event to rollBall EventInstance
		rollBall = FMOD_StudioSystem.instance.GetEvent("event:/rolling");

		// ensures that roll will always have SOME value, i.e. not be null
		if (rollBall.getParameter("rollVel", out rollVel) != FMOD.RESULT.OK){
			Debug.LogError("rollVel parameter not found on rollBall sound event");
			return;
		}
		// begin rollBall EventInstance
		rollBall.start ();
	}

	// use this update for changes that require physics
	void FixedUpdate()
	{
		// grabs player input from the keyboard and transfers to Rigidbody physics component
		float moveHorizontal = Input.GetAxis("Horizontal");
		float moveVertical = Input.GetAxis ("Vertical");

		// report details of H and V movement
		mhText.text = "Horz: " + moveHorizontal.ToString();
		mvText.text = "Vert: " + moveVertical.ToString();

		// new Vector3 variable uses keyboard input for its values: x = horiz key value, y= 0.0f (ball doesn't move up) and z = vert key value
		Vector3 movement = new Vector3 (moveHorizontal, 0.0f, moveVertical);

		// applies input and modulates with speed value; Time.deltaTime smoothes the whole process
		rigidbody.AddForce (movement * speed * Time.deltaTime);

		// get velocity value (magnitude converts to a floating point number)
		pVel = gameObject.rigidbody.velocity.magnitude;

		// report current velocity. This value lets you know what range to use for FMOD event parameters related to ball movement and speed/velocity
		velRollingText.text = "Rolling at: " + pVel.ToString ();

		// use velocity to set volume of rolling and x-fade position for various roll sounds
		rollVel.setValue(pVel);
	}



	void OnTriggerEnter (Collider other) 
	{
		// if the tag matches "PickUp", make it inactive
		// tags must be set in the Unity editor Tag List; tag is applied to the Prefab so that all instances are tagged identically
		if(other.gameObject.tag == "PickUp")
		{
			// hides the cube
			other.gameObject.SetActive(false);
			// increments the count by 1
			count++;
			// function updates the count text field 
			SetCountText ();

			// play a fixed pickup sound: use either this or event:/pickup-ran
			//FMOD_StudioSystem.instance.PlayOneShot("event:/pickup-single", new Vector3(0,0,0));

			// play a randomized pickup sound: use either this or event:/pickup-single
			FMOD_StudioSystem.instance.PlayOneShot("event:/pickup-ran", new Vector3(0,0,0));
		}

		// same as tag comparison for "PickUp"
		if(other.gameObject.tag == "Wall")
		{	
			// reports velocity on moment of impact with a wall
			velText.text = "Velocity: " + pVel.ToString();
			// play a single collide sound
			// NOTE: to do this, each wall contains a duplicate object that has "Is Trigger" checked
			//       Unity will not allow objects with Is Trigger to retain their collision physics     
			//FMOD_StudioSystem.instance.PlayOneShot("event:/pickup-hit", new Vector3(0,0,0));

			// play a collide sound based on velocity
			// this uses sndVel as a parameter in the FMOD timeline of the wall-hit-vel event
			wallHit = FMOD_StudioSystem.instance.GetEvent("event:/wall-hit-vel");

			// ensures that sndVel will always have SOME value, i.e. not be null
			if (wallHit.getParameter("sndVel", out sndVel) != FMOD.RESULT.OK){
				Debug.LogError("sndVel parameter not found on wallHit event");
				return;
			}

			// retrieve current velocity and apply to FMOD parameter sndVel (this chooses the right sound given a velocity value)
			sndVel.setValue(pVel);
			// play the sound
			wallHit.start();
		}
	}

	// function to update the counter text
	void SetCountText ()
	{
		// "prints" new value to the GUI text object
		countText.text = "Count: " + count.ToString();
		// if all 12 PickUps have been collected, report a win message
		if (count >= 12) 
		{
			winText.text = "YOU WIN!";
			// play a victory cue after all 12 cubes are collected
			FMOD_StudioSystem.instance.PlayOneShot("event:/win", new Vector3(0,0,0));
		}
	}
}